/* sgmlmode.c
 *
 * Copyright (c) 1999 Chris Smith
 *
 * Author:
 *  Eckehard Berns  <eb@berns.i-s-o.net>
 *
 * This file is distributed under the GPL, version 2 or at your option any
 * later version.  See COPYING for details.
 */

#include "options.h"

#include SLANG_H
#include <string.h>
#include <ctype.h>

#include "messages.h"
#include "lpecomm.h"
#include "input.h"

#ifdef OPT_SYNTAX_HIGHLIGHTING

enum _states
{
    STATE_BEGIN = 0,
    STATE_TAG,
    STATE_ARGUMENT,
    STATE_QUOTE,
    STATE_VALUE,
    STATE_COMMENT,
    STATE_SLASH
};

#define COLOR_TEXT          0
#define COLOR_TEXT_FG       "lightgray"
#define COLOR_TEXT_BG       "black"

#define COLOR_TAG           1
#define COLOR_TAG_FG        "cyan"
#define COLOR_TAG_BG        "black"

#define COLOR_COMMENT       2
#define COLOR_COMMENT_FG    "green"
#define COLOR_COMMENT_BG    "black"

#define COLOR_SPECIAL       3
#define COLOR_SPECIAL_FG    "brown"
#define COLOR_SPECIAL_BG    "black"

#define COLOR_ARGUMENT      4
#define COLOR_ARGUMENT_FG   "blue"
#define COLOR_ARGUMENT_BG   "black"

#define COLOR_VALUE         5
#define COLOR_VALUE_FG      "brightblue"
#define COLOR_VALUE_BG      "black"

#define COLOR_SLASH_TEXT    6
#define COLOR_SLASH_TEXT_FG "magenta"
#define COLOR_SLASH_TEXT_BG "black"

#define COLOR_SYMBOL        7
#define COLOR_SYMBOL_FG     "white"
#define COLOR_SYMBOL_BG     "black"

#define COLOR_BRACE         8
#define COLOR_BRACE_FG      "yellow"
#define COLOR_BRACE_BG      "black"

#define COLOR_ILLEGAL       9
#define COLOR_ILLEGAL_FG    "brightred"
#define COLOR_ILLEGAL_BG    "black"

#endif /* defined(OPT_SYNTAX_HIGHLIGHTING) */

int mode_accept(buffer *buf)
{
    char *suffix;

    suffix = strrchr(buf->name, '.');
    if (suffix == NULL)
	return 0;
    if (strcmp(suffix, ".sgml") == 0)
	return 1;
    if (strcmp(suffix, ".sgm") == 0)
	return 1;
    if (strcmp(suffix, ".dtd") == 0)
	return 1;

    return 0;
}

void mode_init(buffer * buf)
{
    if (buf->mode_name == NULL)
    {
        buf->hard_tab = 1;
#ifdef OPT_AUTO_INDENT
        buf->autoindent = 0;
#endif /* defined(OPT_AUTO_INDENT) */
    }

    buf->mode_name = "<sgmlmode>";

#ifdef OPT_SYNTAX_HIGHLIGHTING
    buf->state_valid = buf->text;
    buf->state_valid_num = 0;
    buf->text->start_state = STATE_BEGIN;
#endif /* defined(OPT_SYNTAX_HIGHLIGHTING) */
}

void mode_enter(buffer *buf)
{
#ifdef OPT_SYNTAX_HIGHLIGHTING
    SLtt_set_color(COLOR_TEXT, NULL, COLOR_TEXT_FG, COLOR_TEXT_BG);
    SLtt_set_color(COLOR_TAG, NULL, COLOR_TAG_FG, COLOR_TAG_BG);
    SLtt_set_color(COLOR_COMMENT, NULL, COLOR_COMMENT_FG, COLOR_COMMENT_BG);
    SLtt_set_color(COLOR_SPECIAL, NULL, COLOR_SPECIAL_FG, COLOR_SPECIAL_BG);
    SLtt_set_color(COLOR_ARGUMENT, NULL, COLOR_ARGUMENT_FG, COLOR_ARGUMENT_BG);
    SLtt_set_color(COLOR_VALUE, NULL, COLOR_VALUE_FG, COLOR_VALUE_BG);
    SLtt_set_color(COLOR_SLASH_TEXT, NULL, COLOR_SLASH_TEXT_FG, COLOR_SLASH_TEXT_BG);
    SLtt_set_color(COLOR_SYMBOL, NULL, COLOR_SYMBOL_FG, COLOR_SYMBOL_BG);
    SLtt_set_color(COLOR_BRACE, NULL, COLOR_BRACE_FG, COLOR_BRACE_BG);
    SLtt_set_color(COLOR_ILLEGAL, NULL, COLOR_ILLEGAL_FG, COLOR_ILLEGAL_BG);
#endif /* defined(OPT_SYNTAX_HIGHLIGHTING) */
}

#ifdef OPT_FLASH_BRACE

/* SGML implementation of brace flashing
 *
 * Rules:
 * 1. Braces are <>, and they cannot be nested
 * 2. Ignore anything inside quotation marks (double)
 * 3. In SGML mode also use <// as braces
 */
int mode_flashbrace(buffer * buf)
{
    int found;
    int found_slash;
    char ch, quote;

    if (buf->pos.col == 0)
	return 0;

    ch = buf->pos.line->txt[buf->pos.col - 1];
    if ((ch != '>') && (ch != '/'))
	return 0;
    found_slash = (ch == '/') ? 0 : 2;

    buf->pos.col--;
    quote = '\0';
    found = 0;

    do
    {
	while (buf->pos.col <= 0)
	{
	    if (buf->pos.line == buf->scrollpos)
		return 0;
	    buf->pos.line = buf->pos.line->prev;
	    buf->linenum--;
	    buf->pos.col = strlen(buf->pos.line->txt);
	}

	buf->pos.col--;
	ch = buf->pos.line->txt[buf->pos.col];

	if (quote != '\0')
	{
	    if (ch == quote)
		quote = '\0';
	    continue;
	}
	switch (ch)
	{
	case '/':
	    if (found_slash != 2)
	    {
		if (found_slash)
		    return 0;
		found_slash = 1;
	    }
	    break;
	case '<':
	    found = 1;
	    break;

	case '\"':
	    quote = ch;
	    break;
	}
    }
    while (!found);

    set_scr_col(buf);
    return 1;
}

#endif /* defined(OPT_FLASH_BRACE) */

#ifdef OPT_SYNTAX_HIGHLIGHTING

#define IS_NAME(a)   (isalnum(a)   || ((a) == '_') || \
                      ((a) == '%') || ((a) == '&') || \
                      ((a) == '.') || ((a) == '#'))

#define IS_WHITE(a)  (isspace(a))

#define STATE        (*state & 0x00ff)
#define EXCL         (*state & 0xff00)

int mode_highlight(buffer * buf, buf_line * ln, int lnum, int *idx, int *state)
{
    int color;
    int i, ch;
    char *p;

    if (*state == -1)
    {
	*state = buf->state_valid->start_state;
	while (buf->state_valid_num < lnum)
	{
	    i = 0;
	    while (buf->state_valid->txt[i])
		mode_highlight(buf, buf->state_valid,
			       buf->state_valid_num,
			       &i, state);
	    buf->state_valid = buf->state_valid->next;
	    buf->state_valid_num++;
	    buf->state_valid->start_state = *state;
	}
	i = 0;
	color = -1;
	*state = ln->start_state;
	while (i < *idx)
	    color = mode_highlight(buf, ln, lnum, &i, state);
	if ((i > *idx) && (color != -1))
	{
	    *idx = i;
	    return color;
	}
    }

    ch = ln->txt[*idx];
    if (ch == 0)
	return COLOR_TEXT;

    if (ch == '>')
    {
	(*idx)++;
	if (STATE == STATE_BEGIN)
	    return COLOR_ILLEGAL;
	*state = STATE_BEGIN | EXCL;
	return COLOR_TAG;
    }
    if ((*state == (STATE_TAG | 0x100)) &&
        (ch == '-') && (ln->txt[*idx + 1] == '-'))
    {
        *state = STATE_COMMENT | EXCL;
        (*idx) += 2;
    }
    if ((STATE == STATE_TAG) && (IS_WHITE(ch)))
    {
	do
	{
	    (*idx)++;
	}
	while (IS_WHITE(ln->txt[*idx]));
	return COLOR_TAG;
    }
    if (((STATE == STATE_TAG) || (STATE == STATE_ARGUMENT)) &&
        (strchr("-;|+*?,", ch)))
    {
        (*idx)++;
        *state = STATE_TAG | EXCL;
        return COLOR_SYMBOL;
    }
    if (((STATE == STATE_TAG) || (STATE == STATE_ARGUMENT)) &&
        (strchr("()[]", ch)))
    {
        (*idx)++;
        *state = STATE_TAG | EXCL;
        return COLOR_BRACE;
    }
    if ((STATE == STATE_TAG) && (IS_NAME(ch)))
    {
	do
	{
	    (*idx)++;
	}
	while (IS_NAME(ln->txt[*idx]));
	*state = STATE_ARGUMENT | EXCL;
	return COLOR_ARGUMENT;
    }
    if ((STATE == STATE_TAG) && (ch == '"'))
    {
	(*idx)++;
	*state = STATE_QUOTE | EXCL;
    }
    if ((STATE == STATE_TAG) && (ch == '/'))
    {
	(*idx)++;
	*state = STATE_SLASH | EXCL;
	return COLOR_TAG;
    }
    if (STATE == STATE_TAG)
    {
	(*idx)++;
	return COLOR_ILLEGAL;
    }
    if ((STATE == STATE_ARGUMENT) && (ch == '='))
    {
	(*idx)++;
	if (ln->txt[*idx] == '"')
	{
	    *state = STATE_TAG | EXCL;
	}
	else
	{
	    *state = STATE_VALUE | EXCL;
	}
	return COLOR_TAG;
    }
    if (STATE == STATE_ARGUMENT)
    {
	if (IS_WHITE(ch))
	{
	    (*idx)++;
	    *state = STATE_TAG | EXCL;
	    return COLOR_TAG;
	}
	(*idx)++;
	return COLOR_ILLEGAL;
    }
    if (STATE == STATE_VALUE)
    {
	while ((!IS_WHITE(ln->txt[*idx])) && (ln->txt[*idx]) &&
	       (ln->txt[*idx] != '>'))
	    (*idx)++;
	*state = STATE_TAG | EXCL;
	return COLOR_VALUE;
    }
    if (STATE == STATE_QUOTE)
    {
	while ((ln->txt[*idx]) && (ln->txt[*idx] != '"'))
	    (*idx)++;
	if (ln->txt[*idx] == '"')
	{
	    (*idx)++;
	    *state = STATE_TAG | EXCL;
	}
	return COLOR_VALUE;
    }
    if (STATE == STATE_COMMENT)
    {
	p = strstr(&(ln->txt[*idx]), "--");
	if (p == NULL)
	{
	    *idx = strlen(ln->txt);
	}
	else
	{
	    *state = STATE_TAG | EXCL;
	    *idx = p + 2 - ln->txt;
	}
	return COLOR_COMMENT;
    }
    if (*state == STATE_SLASH)
    {
	if (ch == '/')
	{
	    (*idx)++;
	    *state = STATE_BEGIN;
	    return COLOR_TAG;
	}
	p = strchr(&(ln->txt[*idx]), '/');
	if (p == NULL)
	    p = &(ln->txt[strlen(ln->txt)]);
	color = COLOR_SLASH_TEXT;
	*idx = p - ln->txt;
	return color;
    }
    if (ch == '<')
    {
	*state = STATE_TAG;
	(*idx)++;
        if (ln->txt[*idx] == '!')
        {
            *state = STATE_TAG | 0x100;
            (*idx)++;
        }
	else if (ln->txt[*idx] == '/')
        {
	    (*idx)++;
        }
	ch = ln->txt[*idx];
	while (IS_NAME(ch))
	{
	    (*idx)++;
	    ch = ln->txt[*idx];
	}
	if (ch == '/')
	{
	    *state = STATE_SLASH;
	    (*idx)++;
	}
	return COLOR_TAG;
    }
    if (ch == '&')
    {
	do
	{
	    (*idx)++;
	    ch = ln->txt[*idx];
	}
	while ((IS_NAME(ch)) || (ch == '#'));
	if (ch == ';')
	{
	    (*idx)++;
	    return COLOR_SPECIAL;
	}
	return COLOR_ILLEGAL;
    }

    *idx += strcspn(&(ln->txt[*idx]), "<&");
    return COLOR_TEXT;
}

#endif /* defined(OPT_SYNTAX_HIGHLIGHTING) */
